/*
 * Copyright (C) 2014 Erik-Oliver Blass
 *                    Travis Mayberry
 *                    Guevara Noubir
 *                    Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * dm-hive.c, ver.2014.11.03
 *
 * HiVE device-mapper target.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/bio.h>
#include <linux/slab.h>
#include <linux/device-mapper.h>
#include <linux/dm-io.h>
#include <linux/semaphore.h>
#include <linux/crypto.h>
#include <linux/scatterlist.h>
#include <crypto/rng.h>
#include <linux/workqueue.h>
#include <linux/list.h>
#include <linux/kfifo.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

/*
 * Enables the use of simple stupid random sector selection, instead of using a
 * knuth shuffle. The dumb method is much more efficient with large drives and
 * small K, so we go with it.
 */
#define HIVE_DUMB_RANDOM_SECTORS

#define DM_MSG_PREFIX "hive"

#define HIVE_K 3           /* K extra writes. */

#define HIVE_SECTOR 4096       /* In bytes. */
#define HIVE_HW_SECTOR 512     /* In bytes. */
#define HIVE_SECTOR_SCALE (HIVE_SECTOR / HIVE_HW_SECTOR)

#define HIVE_HEADERS_LEN 1 /* In blocks. */
#define HIVE_KEY_LEN 32    /* In bytes. */
#define HIVE_IV_LEN 16     /* In bytes. */
#define HIVE_STASH_LEN 64  /* In sectors. */
#define HIVE_NAME_LEN 16   /* HiVE instance name, in bytes. */
#define HIVE_SALT_LEN 32   /* In bytes. */
#define HIVE_DIGEST_LEN 32 /* In bytes. */

/* Crypto operations. */
#define HIVE_ENCRYPT 1
#define HIVE_DECRYPT 2

#define HIVE_MAX_WORLDS 2  /* 2 worlds supported in each HiVE. */
#define HIVE_WORLD_0 0     /* Public world. */
#define HIVE_WORLD_1 1     /* Hidden world. */
#define HIVE_INVALID_WORLD 0xff

#define HIVE_META_PER_SECTOR 128
#define HIVE_MAP_PER_SECTOR 512

/* Cache flags & constants. */
#define HIVE_DIRTY       0x000000001
#define HIVE_CACHE_HINT  0x000000002
#define HIVE_EVICT_ALL 1

/* Stash constants. */
#define HIVE_PAGE_TO_STASH 1
#define HIVE_STASH_TO_PAGE 2
#define HIVE_INVALID_STASH  0xffffffffffffffff
#define HIVE_INVALID_SECTOR 0xffffffffffffffff

/*
 * Memory pools.
 * TODO: These are quite large, could be reduced later after a
 * proper analysis of the actual requirements.
 */
#define HIVE_BIOSET_SIZE 64
#define HIVE_PAGEPOOL_SIZE 128
#define HIVE_IO_WORKPOOL_SIZE 128
#define HIVE_CRYPTO_WORKPOOL_SIZE 128
#define HIVE_METACACHEPOOL_SIZE 32
#define HIVE_MAPCACHEPOOL_SIZE 32
#define HIVE_STASHPOOL_SIZE 128

#define HIVE_MAP_CACHE_FULL HIVE_MAPCACHEPOOL_SIZE

/* Return codes. */
#define HIVE_SUCCESS 0
#define HIVE_ERROR 1

/* /proc file listing mapped HiVE devices. */
#define HIVE_PROC_FILE "hivetab"

#define HIVE_BIO_FLAGS 0

/*
 * Random data generation algorithm.
 */
#define HIVE_PRNG_AESCTR 0
#define HIVE_PRNG_ARC4 1

#define HIVE_PRNG HIVE_PRNG_AESCTR /* Pick one of the above. */

#if (HIVE_PRNG == HIVE_PRNG_AESCTR)
/* Refresh keys after generating this much data. */
#define HIVE_PRNG_AESCTR_REFRESH_LEN 1048576 /* In bytes. */ /* 1 MB */
#define HIVE_PRNG_AESCTR_CHUNK_LEN 1  /* No of chunks to generate the data in. */

/*
 * AES-CTR context for random data generation.
 */
struct hive_rand_context {
	u8 *buf;
	u64 max_chunk;
	u64 cur_chunk;
	u64 max_byte;
	u64 cur_byte;
	struct crypto_blkcipher *tfm;
};

#elif (HIVE_PRNG == HIVE_PRNG_ARC4)
/*
 * ARC4 context for random data generation.
 */
struct hive_rand_context {
	u8 S[256];
	u8 i;
	u8 j;
};

#else
#error "PRNG type not defined in dm-hive.c"
#endif

struct hive_dev;

/* Shared resources for a HiVE instance. */
struct hive_resource {
	u64 start;         /* HiVE starts here on real device. */
	u64 data_start;    /* Data sectors start here. */
	u64 data_len;
	u64 meta_start;    /* Meta sectors start here. */
	u32 is_single_world;

#ifdef HIVE_DUMB_RANDOM_SECTORS
	u64 *random_list; /* Only if using the smart & inefficient random sector
			   * selection. */
#endif

	struct hive_rand_context rand; /* ARC4 context. */

	/* Work queues. */
	struct workqueue_struct *io_queue;
	struct workqueue_struct *crypto_queue;

	/* Memory pools. */
	struct bio_set *bioset;
	mempool_t *page_pool;
	struct kmem_cache *_io_work_pool;
	mempool_t *io_work_pool;
	struct kmem_cache *_crypto_work_pool;
	mempool_t *crypto_work_pool;
	struct kmem_cache *_meta_cache_pool;
	mempool_t *meta_cache_pool;
	struct kmem_cache *_map_cache_pool;
	mempool_t *map_cache_pool;
	struct kmem_cache *_stash_pool;
	mempool_t *stash_pool;

	/* Locks. */
	struct semaphore io_lock;
	struct semaphore crypto_lock;

	/* Associated HiVE device. */
	struct hive_dev *dev;

	/* Meta block cache. */
	struct hive_meta_cache *meta_cache[HIVE_METACACHEPOOL_SIZE];
	u64 meta_cache_count;
};

/* Stash entry, contains one sector of data. */
struct hive_stash {
	u64 sector;
	char *data;
};

/* Hive context for a world. */
struct hive_config {
	struct dm_dev *real_dev;  /* Underlying block device. */
	struct dm_dev *map_dev;   /* Map device, if these is one. */
	u64 *map;                 /* In-memory map, if there is one. */
	u64 world;                /* World no. */
	u64 world_len;            /* World size. */
	struct hive_config *worlds[HIVE_MAX_WORLDS]; /* Ref to all configs in this HiVE. */
	struct crypto_blkcipher *tfm;                  /* This is initialized with the world's own key. */
	DECLARE_KFIFO(stash, struct hive_stash *, HIVE_STASH_LEN);
	struct hive_map_cache *map_cache[HIVE_MAPCACHEPOOL_SIZE]; /* Each world gets a separate map cache. */
	u64 map_cache_count;
	struct hive_resource *res; /* Shared resources. */
};

/* Represents a HiVE instance with up to 2 worlds. */
struct hive_dev {
	dev_t dev;
	char hive_name[16]; /* Instance name. */
	char *world_paths[HIVE_MAX_WORLDS];
	u32 worlds;
	struct hive_dev *map_dev;
	struct hive_config *hc[HIVE_MAX_WORLDS];
	struct list_head list;
};
static LIST_HEAD(hive_dev_list); /* We keep all HiVEs in a list. */
static DEFINE_SEMAPHORE(hive_dev_lock);

/* HiVE header entry, one for each world. */
struct hive_header {
	u8 key[HIVE_KEY_LEN];
	u8 key_digest[HIVE_DIGEST_LEN];
	u8 key_salt[HIVE_SALT_LEN];
	u8 pass_salt[HIVE_SALT_LEN];
	u64 stash_start;
	u64 stash_len;
	u64 stash_map_start;
	u64 stash_map_len;
	u64 map_start;
	u64 map_len;
	u64 meta_start;
	u64 data_start;
	u64 data_len;
	u64 world_len;
	u64 next_hive;
	u8 iv[HIVE_IV_LEN];
	u64 padding1;
	u64 padding2;
	u64 padding3; /* Must extend header sector size if more data needed. */
};

/* Represents an IO in flight. */
struct hive_io_work {
	struct hive_config *hc;
	struct bio *bio;         /* Original bio sent to the real device. */
	struct work_struct work;
};

/* Represents a crypto operation in flight. */
struct hive_crypto_work {
	struct hive_config *hc;
	struct bio *bio;         /* Bio of the read chunk to be decrypted. */
	struct hive_io_work *io;
	u8 iv[HIVE_IV_LEN];
	struct work_struct work;
};

/* A meta block, containing the IV & reverse map for a data sector. */
struct hive_meta {
	u8 iv[HIVE_IV_LEN];
	u64 r_map;
	u64 world;
};

/* A cache entity holding all meta blocks read from a sector. */
struct hive_meta_cache {
	u64 sector;
	char *data;
	u8 is_decrypted[HIVE_META_PER_SECTOR / 8]; /* Bitmap. Is the meta block at position X decrypted? */
	u8 who_decrypted[HIVE_META_PER_SECTOR / 8]; /* Bitmap. If so, who decrypted it? */
};

/*
 * A cache entity holding all forward map entries read from a recursive HiVE
 * sector.
 */
struct hive_map_cache {
	u64 sector;
	u64 *map;
	u8 is_accessed[HIVE_MAP_PER_SECTOR / 8]; /* Bitmap. How many entries in the current cache is
						   * accessed? */
	u32 flags;
};

static void hive_set_iv(u64, u8 *, struct hive_config *);
static void hive_write_endio(struct bio *, int);
static void hive_queue_crypto(struct hive_crypto_work *);

/*
 * Miscellaneous helper functions.
 */

/* Bitset manipulation operations. */
static unsigned hive_bit_count(u8 *bitset, u64 len)
{
	unsigned i;
	unsigned j;
	unsigned count;
	u8 tmp;

	count = 0;
	for (i = 0; i < (len / 8); ++i) {
		tmp = bitset[i];
		for (j = 0; j < 8; ++j) {
			count += (tmp & 0x1);
			tmp = tmp >> 1;
		}
	}

	return count;
}

static inline void hive_set_bit(u8 *bitset, unsigned bit)
{
	bitset[bit >> 3] |= (((u8) 0x1) << (bit & 0x7));
}

static inline void hive_clear_bit(u8 *bitset, unsigned bit)
{
	bitset[bit >> 3] &= ~(((u8) 0x1) << (bit & 0x7));
}

static void hive_set_bit_to_value(u8 *bitset, unsigned bit, u8 value)
{
	if (value == 0)
		hive_clear_bit(bitset, bit);
	else
		hive_set_bit(bitset, bit);
}

static inline unsigned hive_test_bit(u8 *bitset, unsigned bit)
{
	return bitset[bit >> 3] & (((u8) 0x1) << (bit & 0x7));
}

static inline void hive_clear_bitset(u8 *bitset, unsigned len)
{
	memset(bitset, 0, len >> 3);
}

/* Decodes a hex encoded byte string. */
static u8 *hive_hex_decode(u8 *hex)
{
	u8 buf[3];
	u8 *s;
	unsigned len;
	unsigned i;

	buf[2] = '\0';
	len = strlen(hex) / 2;

	s = kmalloc(len, GFP_KERNEL);
	memset(s, 0, len);

	for (i = 0; i < len; ++i) {
		buf[0] = *hex++;
		buf[1] = *hex++;
		kstrtou8(buf, 16, &s[i]);
	}

	return s;
}

/*
 * /proc file functions.
 */

/* Iterates over all hive_devs and prints the mapped device info. */
static int hive_list_mounts(struct seq_file *f, void *v)
{
	struct hive_dev *cur;
	unsigned i;

	down(&hive_dev_lock);
	list_for_each_entry(cur, &hive_dev_list, list) {
		seq_printf(f, "%s ", cur->hive_name);

		seq_printf(f, "%u ", cur->worlds);
		for (i = 0; i < cur->worlds; ++i)
			seq_printf(f, "%s ", cur->world_paths[i]);

		if (cur->map_dev) {
			seq_printf(f, "%u ", cur->map_dev->worlds);
			for (i = 0; i < cur->map_dev->worlds; ++i)
				seq_printf(f, "%s ", cur->map_dev->world_paths[i]);
		} else {
			seq_puts(f, "0 ");
		}
		seq_puts(f, "\n");
	}
	up(&hive_dev_lock);

	return 0;
}

/* Open handler for the /proc file. */
static int hive_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, hive_list_mounts, NULL);
}

/* /proc file operations. */
static const struct file_operations hive_fops = {
	.owner = THIS_MODULE,
	.open = hive_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};


/*
 * Memory pool management functions. Pretty self-explanatory.
 */

static struct page *hive_allocate_page(struct hive_config *hc)
{
	struct page *p;

	p = mempool_alloc(hc->res->page_pool, GFP_KERNEL);
	if (!p)
		DMCRIT("Cannot allocate new page!");

	return p;
}

static inline void hive_free_page(struct page *p, struct hive_config *hc)
{
	mempool_free(p, hc->res->page_pool);
}

static void hive_free_sector(u8 *s, struct hive_config *hc)
{
	struct page *p;

	p = virt_to_page(s);
	kunmap(p);
	mempool_free(p, hc->res->page_pool);
}

static struct hive_stash *hive_allocate_stash(struct hive_config *hc)
{
	struct page *p;
	struct hive_stash *hs;

	p = hive_allocate_page(hc);

	hs = mempool_alloc(hc->res->stash_pool, GFP_KERNEL);
	if (!hs)
		DMCRIT("Cannot allocate new stash!");
	else
		hs->data = kmap(p);

	return hs;
}

static void hive_free_stash_keep_page(struct hive_stash *hs, struct hive_config *hc)
{
	struct page *p;

	p = virt_to_page(hs->data);
	kunmap(p);
	mempool_free(hs, hc->res->stash_pool);
}

static void hive_free_stash(struct hive_stash *hs, struct hive_config *hc)
{
	struct page *p;

	p = virt_to_page(hs->data);
	kunmap(p);
	hive_free_page(p, hc);
	mempool_free(hs, hc->res->stash_pool);
}

static struct bio *hive_allocate_bio(struct hive_config *hc)
{
	struct bio *b;

	b = bio_alloc_bioset(GFP_KERNEL, 1, hc->res->bioset);
	if (!b)
		DMCRIT("Cannot allocate new bio!");

	return b;
}

static struct hive_io_work *hive_allocate_io_work(struct bio *bio, struct hive_config *hc)
{
	struct hive_io_work *w;

	w = mempool_alloc(hc->res->io_work_pool, GFP_NOIO);
	if (!w) {
		DMCRIT("Cannot allocate new io work!");
	} else {
		w->bio = bio;
		w->hc = hc;
	}
	return w;
}

static inline void hive_free_io_work(struct hive_io_work *w, struct hive_config *hc)
{
	mempool_free(w, hc->res->io_work_pool);
}

static struct hive_crypto_work *hive_allocate_crypto_work(struct hive_config *hc)
{
	struct hive_crypto_work *w;

	w = mempool_alloc(hc->res->crypto_work_pool, GFP_NOIO);
	if (!w)
		DMCRIT("Cannot allocate new crypto work!");
	else
		w->hc = hc;

	return w;
}

static inline void hive_free_crypto_work(struct hive_crypto_work *w, struct hive_config *hc)
{
	mempool_free(w, hc->res->crypto_work_pool);
}

static struct hive_meta_cache *hive_allocate_meta_cache(struct hive_config *hc)
{
	struct hive_meta_cache *mc;

	mc = mempool_alloc(hc->res->meta_cache_pool, GFP_NOIO);
	if (!mc)
		DMCRIT("Cannot allocate new meta cache!");

	return mc;
}

static inline void hive_free_meta_cache(struct hive_meta_cache *mc, struct hive_config *hc)
{
	mempool_free(mc, hc->res->meta_cache_pool);
}

static struct hive_map_cache *hive_allocate_map_cache(struct hive_config *hc)
{
	struct hive_map_cache *mc;

	mc = mempool_alloc(hc->res->map_cache_pool, GFP_NOIO);
	if (!mc)
		DMCRIT("Cannot allocate new meta cache!");

	return mc;
}

static inline void hive_free_map_cache(struct hive_map_cache *mc, struct hive_config *hc)
{
	mempool_free(mc, hc->res->map_cache_pool);
}

/*
 * HiVE device management functions.
 */

/* Looks up a block device by path. */
static int __hive_lookup_dev(char *dev_path, dev_t *dev)
{
	struct block_device *bdev;

	bdev = lookup_bdev(dev_path);
	if (IS_ERR(bdev))
		return HIVE_ERROR;

	*dev = bdev->bd_dev;
	bdput(bdev);

	return HIVE_SUCCESS;
}

/* Looks up a HiVE device by its name and underlying block device. */
static struct hive_dev *hive_lookup_dev(char *dev_path, char *name)
{
	struct hive_dev *cur;
	dev_t dev;

	if (__hive_lookup_dev(dev_path, &dev) == HIVE_ERROR) {
		DMCRIT("Device lookup failed!");
		return NULL;
	}

	list_for_each_entry(cur, &hive_dev_list, list) {
		if (cur->dev == dev && strncmp(cur->hive_name, name, HIVE_NAME_LEN) == 0)
			return cur;
	}

	return NULL;
}

/* Creates a new HiVE device. */
static struct hive_dev *hive_create_dev(char *dev_path, char *name, u32 is_single_world)
{
	unsigned i;
	unsigned max_worlds = HIVE_MAX_WORLDS;
	struct hive_dev *hd = kmalloc(sizeof(*hd), GFP_KERNEL);

	if (__hive_lookup_dev(dev_path, &hd->dev) == HIVE_ERROR) {
		DMCRIT("Device lookup failed!");
		return NULL;
	}
	strncpy(hd->hive_name, name, HIVE_NAME_LEN);
	hd->worlds = 0;

	if (is_single_world)
		max_worlds = 1;

	/* Also allocate the contexts for both worlds here. */
	for (i = 0; i < max_worlds; ++i) {
		hd->hc[i] = kmalloc(sizeof(struct hive_config), GFP_KERNEL);
		memset(hd->hc[i], 0, sizeof(struct hive_config));
		hd->world_paths[i] = NULL;
	}

	INIT_LIST_HEAD(&hd->list);
	list_add(&hd->list, &hive_dev_list);

	return hd;
}

/* Destroys a HiVE device. */
static void hive_destroy_dev(struct hive_dev *hd)
{
	unsigned i;
	unsigned max_worlds = 2;

	if (hd->hc[0]->res->is_single_world)
		max_worlds = 1;

	for (i = 0; i < max_worlds; ++i) {
		kfree(hd->hc[i]);
		kfree(hd->world_paths[i]);
	}

	list_del(&hd->list);
}

/* Reduces the reference count on a HiVE device and destroy when zero. */
static inline void hive_put_dev(struct hive_dev *hd)
{
	if (!--hd->worlds)
		hive_destroy_dev(hd);
}

/*
 * Checks if there is unmounted worlds on the given HiVE, and increases its
 * reference count if so.
 */
static int hive_get_dev(struct hive_dev *hd, u32 *world, struct dm_target *ti)
{
	if (hd->worlds == 1 && hd->hc[0]->res->is_single_world) {
		ti->error = "HIVE operating in single volume mode";
		return HIVE_ERROR;
	}

	if (hd->worlds == 2) {
		ti->error = "HiVE device already has 2 volumes active";
		return HIVE_ERROR;
	}

	*world = hd->worlds++;

	return HIVE_SUCCESS;
}

/* Sets the world's mapped device path. */
static void hive_set_world_path(char *path, struct hive_dev *hd)
{
	hd->world_paths[hd->worlds - 1] = kmalloc(strlen(path) + 1, GFP_KERNEL);
	strcpy(hd->world_paths[hd->worlds - 1], path);
}


/*
 * Disk I/O helpers.
 */

/*
 * Sync I/O helper for metadata operations.
 * @bdev: Block device.
 * @sector: Sector no on REAL device!
 * @rw: READ or WRITE
 * @write_buf: Data buffer to write. Ignored for reads.
 * @return: Buffer containing read data. Caller frees memory.
*/

static void *__hive_rw_sector(struct block_device *bdev, u64 sector,
			int rw, void *write_buf, struct hive_config *hc)
{
	struct bio *bio;
	struct page *p;

	if (rw == WRITE && !write_buf) {
		DMCRIT("Write buffer is NULL, aborting");
		return NULL;
	}

	bio = hive_allocate_bio(hc);
	bio->bi_bdev = bdev;
	bio->bi_sector = sector * HIVE_SECTOR_SCALE;

	if (rw == READ) {
		p = hive_allocate_page(hc);
		bio->bi_rw &= ~REQ_WRITE;
	} else {
		p = virt_to_page(write_buf);
		bio->bi_rw |= REQ_WRITE;
	}

	bio_add_page(bio, p, HIVE_SECTOR, 0);

	submit_bio_wait(HIVE_BIO_FLAGS, bio);

	bio_put(bio);

	return kmap(p);
}

/* Shortcut for I/O on the underlying block device. */
static inline void *hive_rw_sector(u64 sector, int rw, void *write_buf, struct hive_config *hc)
{
	return __hive_rw_sector(hc->worlds[0]->real_dev->bdev, sector, rw, write_buf, hc);
}

/* Shortcut for I/O on the map HiVE. */
static inline void *hive_rw_map_sector(u64 sector, int rw, void *write_buf, struct hive_config *hc)
{
	return __hive_rw_sector(hc->map_dev->bdev, sector, rw, write_buf, hc);
}


/*
 * Random data helpers.
 */

/* Returns crypto-safe random bytes from kernel pool. */
static inline void hive_get_random_bytes_kernel(u8 *data, u64 len)
{
	crypto_get_default_rng();
	crypto_rng_get_bytes(crypto_default_rng, data, len);
	crypto_put_default_rng();
}

#if (HIVE_PRNG == HIVE_PRNG_AESCTR)
/* Sets a new random AES-CTR key if necessary and refreshes the random data
 * buffer. */
static void hive_fill_rand_buf(struct hive_rand_context *rand)
{
	u8 key[32];
	struct scatterlist src;
	struct scatterlist dst;
	struct blkcipher_desc desc;

	/* Refresh the key. */
	if (rand->cur_chunk == rand->max_chunk) {
		hive_get_random_bytes_kernel(key, 32);
		crypto_blkcipher_setkey(rand->tfm, key, 32);
		memset(key, 0, 32);

		rand->cur_chunk = 0;
	}

	sg_init_table(&src, 1);
	sg_init_table(&dst, 1);

	sg_set_buf(&src, rand->buf, rand->max_byte);
	sg_set_buf(&dst, rand->buf, rand->max_byte);

	desc.tfm = rand->tfm;
	desc.flags = 0;

	if (crypto_blkcipher_encrypt(&desc, &dst, &src, rand->max_byte))
		DMCRIT("Error generating random stream");

	++(rand->cur_chunk);
	rand->cur_byte = 0;
}

/* Generates random bytes using the passed AES-CTR context. */
static void hive_get_random_bytes(u8 *data, u64 len, struct hive_rand_context *rand)
{
	u64 left;
	u64 read;

	while (len) {
		left = rand->max_byte - rand->cur_byte;
		if (!left)
			hive_fill_rand_buf(rand);

		read = (left < len) ? left : len;

		memcpy(data, rand->buf + rand->cur_byte, read);
		data += read;
		rand->cur_byte += read;
		len -= read;
	}
}

#elif (HIVE_PRNG == HIVE_PRNG_ARC4)
/* Generates random bytes using the passed ARC4 context. */
static void hive_get_random_bytes(u8 *data, u64 len, struct hive_rand_context *rand)
{
	u8 *S;
	unsigned i;
	unsigned j;
	unsigned k;

	S = rand->S;
	i = rand->i;
	j = rand->j;

	for (k = 0; k < len; ++k) {
		i = (i + 1) & 0xff;
		j = (j + S[i]) & 0xff;
		swap(S[i], S[j]);
		data[k] = S[(S[i] + S[j]) & 0xff];
	}

	rand->i = i;
	rand->j = j;
}
#endif

/* Returns random 64bit value using random context. */
static inline u64 hive_get_random_u64(struct hive_config *hc)
{
	u64 r;
	hive_get_random_bytes((u8 *)&r, 8, &hc->res->rand);
	return r;
}

/* Fills a buffer with a random IV. */
static inline void hive_get_random_iv(u8 *iv, struct hive_config *hc)
{
	hive_get_random_bytes(iv, HIVE_IV_LEN, &hc->res->rand);
}

/* Fills a sector with random bytes. */
static inline void hive_fill_random_sector(unsigned char *data, struct hive_config *hc)
{
	hive_get_random_bytes(data, HIVE_SECTOR, &hc->res->rand);
}

/* Fills a meta block with random bytes. */
static inline void hive_fill_random_meta(struct hive_meta *m, struct hive_config *hc)
{
	hive_get_random_bytes((char *) m, sizeof(*m), &hc->res->rand);
}

#ifdef HIVE_DUMB_RANDOM_SECTORS
/*
 * Returns K random sector numbers. Dumb way of doing it, which happens to be
 * better than all alternatives for this application -- we have a very large
 * domain, but a very small K.
 */
static void hive_get_random_sectors(u64 count, u64 *out, struct hive_config *hc)
{
	u64 i;
	u64 k;
	u64 r;
	unsigned is_unique;

	for (k = 0; k < count;) {
		is_unique = 1;
		r = hive_get_random_u64(hc) % hc->res->data_len;

		for (i = 0; i < k; ++i) {
			if (r == out[i]) {
				is_unique = 0;
				break;
			}
		}

		if (is_unique)
			out[k++] = r;
	}
}
#else
/*
 * Returns K random sector numbers. Essentially a partial Knuth shuffle in
 * reverse.
 */
static void hive_get_random_sectors(u64 count, u64 *out, struct hive_config *hc)
{
	u64 i;
	u64 j;
	u64 k;

	k = 0;
	for (i = 0; k < count; ++i) {
		j = i + (hive_get_random_u64(hc) % (hc->res->data_len - i));
		swap(hc->res->random_list[i], hc->res->random_list[j]);
		out[k++] = hc->res->random_list[i];
	}
}
#endif

/*
 * Crypto functions.
 */

/* Convert one buffer of data. */
static void __hive_do_crypto(struct scatterlist *src, struct scatterlist *dst,
			u64 len, u8 *iv, int op, struct hive_config *hc)
{
	struct blkcipher_desc desc;

	desc.tfm = hc->tfm;
	desc.flags = 0;

	/*
	 * Lock so that another crypto work doesn't change the IV before we
	 * finish our job.
	 */
	down(&hc->res->crypto_lock);
	if (iv)
		crypto_blkcipher_set_iv(hc->tfm, iv, HIVE_IV_LEN);

	if (op == HIVE_ENCRYPT) {
		if (crypto_blkcipher_encrypt(&desc, dst, src, len))
			DMCRIT("Error doing crypto");
	} else if (op == HIVE_DECRYPT) {
		if (crypto_blkcipher_decrypt(&desc, dst, src, len))
			DMCRIT("Error doing crypto");
	} else {
		DMCRIT("Unknown crypto operation");
	}
	up(&hc->res->crypto_lock);
}

/* Convert a data buffer. */
static void hive_do_crypto_from_buffer(char *buf, u64 len, u8 *iv, int op, struct hive_config *hc)
{
	struct scatterlist src;
	struct scatterlist dst;

	sg_init_table(&src, 1);
	sg_init_table(&dst, 1);

	sg_set_buf(&src, buf, len);
	sg_set_buf(&dst, buf, len);

	__hive_do_crypto(&src, &dst, len, iv, op, hc);
}

/* Convert a page. */
static void hive_do_crypto_from_page(struct page *p, unsigned offset,
				u64 len, u8 *iv, int op, struct hive_config *hc)
{
	char *buf = ((char *) kmap(p)) + offset;

	hive_do_crypto_from_buffer(buf, len, iv, op, hc);

	kunmap(p);
}

/*
 * Convert a buffer with a random IV, then adjust the corresponding sector's
 * meta block with the new IV.
 */
static void hive_encrypt_sector_random_iv(u64 sector, char *buf, struct hive_config *hc)
{
	u8 iv[HIVE_IV_LEN];

	hive_get_random_iv(iv, hc);
	hive_do_crypto_from_buffer(buf, HIVE_SECTOR, iv, HIVE_ENCRYPT, hc);
	hive_set_iv(sector, iv, hc);
}

/* Decrypt a page. */
static inline void hive_decrypt_sector_page(struct page *p, unsigned offset,
					u8 *iv, struct hive_config *hc)
{
	hive_do_crypto_from_page(p, offset, HIVE_SECTOR, iv, HIVE_DECRYPT, hc);
}

/* Decrypt a sector. */
static inline void hive_decrypt_sector_buffer(char *buf, u8 *iv, struct hive_config *hc)
{
	hive_do_crypto_from_buffer(buf, HIVE_SECTOR, iv, HIVE_DECRYPT, hc);
}

/*
 * Forward Map functions.
 */

/* Returns the number of maps entries actually accessed in a map cache. */
static inline unsigned hive_accessed_map_count(struct hive_map_cache *mc)
{
	return hive_bit_count(mc->is_accessed, HIVE_MAP_PER_SECTOR);
}

/* Writes the accessed map entries back to disk. */
static void __hive_evict_map_cache(u32 op, struct hive_config *hc)
{
	u64 i;
	u64 j;
	u64 write_count;
	u64 free_i = 0;

	/*
	 * If this world has an in memory map, this function should't have been
	 * called.
	 */
	if (hc->map != NULL)
		return;

	for (i = 0; i < hc->map_cache_count; ++i) {
		/*
		 * We have a mechanism to keep certain map entries with 'cache
		 * hints' in the cache as long as there is room in the
		 * cache. This functionality is disabled at the moment until we
		 * properly analyze the security implications of this
		 * approach.
		 */
		if (op != HIVE_EVICT_ALL && (hc->map_cache[i]->flags & HIVE_CACHE_HINT)) {
				hc->map_cache[free_i++] = hc->map_cache[i];
		} else {
			/*
			 * Only write back the sector if a cache entry in that
			 * sector is modified.
			 */
			if (hc->map_cache[i]->flags & HIVE_DIRTY) {
				write_count = hive_accessed_map_count(hc->map_cache[i]);

				/*
				 * If multiple map entries in the same sector
				 * has been accessed, we need to write the
				 * sector back multiple times as well to ensure
				 * that K blocks are modified in the map HiVE
				 * for each access.
				 */
				for (j = 0; j < write_count; ++j) {
					hive_rw_map_sector(hc->map_cache[i]->sector, WRITE,
							(char *) hc->map_cache[i]->map, hc);
				}

			}
			hive_free_sector((char *) hc->map_cache[i]->map, hc);
			hive_free_map_cache(hc->map_cache[i], hc);
		}
	}

	hc->map_cache_count = free_i;
}

/* Evicts all map caches, also those with cache hints. */
static inline void hive_evict_map_cache_all(struct hive_config *hc)
{
	__hive_evict_map_cache(HIVE_EVICT_ALL, hc->worlds[HIVE_WORLD_0]);
	if (!hc->res->is_single_world)
		__hive_evict_map_cache(HIVE_EVICT_ALL, hc->worlds[HIVE_WORLD_1]);
}

/*
 * This would normally only evict the map entries without cache hints. This is
 * disabled at the moment until we evaluate the security implications of caching
 * map blocks across multiple I/O requests.
 */
static inline void hive_evict_map_cache(struct hive_config *hc)
{
	hive_evict_map_cache_all(hc);
	/* __hive_evict_map_cache(0, hc); */ /* Disabled.*/
}

/* Finds and returns a cached map entry. */
static struct hive_map_cache *hive_find_map_cache(u64 sector, struct hive_config *hc)
{
	u64 i;

	for (i = 0; i < hc->map_cache_count; ++i) {
		if (hc->map_cache[i]->sector == sector)
			return hc->map_cache[i];
	}

	return NULL;
}

/* Caches a newly read map HiVE sector. */
static struct hive_map_cache *hive_cache_map(u64 sector, char *data, struct hive_config *hc)
{
	struct hive_map_cache *mc;

	/* This should never happen. */
	if (unlikely(hc->map_cache_count == HIVE_MAP_CACHE_FULL)) {
		DMCRIT("Cache full, evicting!");
		hive_evict_map_cache_all(hc);
	}

	mc = hive_allocate_map_cache(hc);
	mc->sector = sector;
	mc->map = (u64 *) data;
	mc->flags = 0;
	hive_clear_bitset(mc->is_accessed, HIVE_MAP_PER_SECTOR);

	hc->map_cache[hc->map_cache_count++] = mc;

	return mc;
}

/*
 * Returns a map cache. It the entry has not been cached yet, it is first read
 * from the map HiVE and cached.
 */
static struct hive_map_cache *hive_get_map_cache(u64 sector, struct hive_config *hc, u64 *index)
{
	struct hive_map_cache *mc;
	u64 map_dev_sector;
	char *data;

	if (hc->map != NULL)
		return NULL;

	map_dev_sector = sector / HIVE_MAP_PER_SECTOR;
	*index = sector % HIVE_MAP_PER_SECTOR;
	mc = hive_find_map_cache(map_dev_sector, hc);

	if (!mc) {
		data = hive_rw_map_sector(map_dev_sector, READ, NULL, hc);
		mc = hive_cache_map(map_dev_sector, data, hc);
		hive_set_bit(mc->is_accessed, *index);
	} else {
		if (!hive_test_bit(mc->is_accessed, *index)) {
			data = hive_rw_map_sector(map_dev_sector, READ, NULL, hc);
			hive_free_sector(data, hc);
			hive_set_bit(mc->is_accessed, *index);
		}
	}

	return mc;
}

/* Returns the forward map for the given sector. */
static u64 __hive_map_sector(u64 sector, u32 flags, struct hive_config *hc)
{
	struct hive_map_cache *mc;
	u64 index;

	/* If map is in memory, return it immediately. */
	if (hc->map)
		return hc->map[sector];

	/* Otherwise, map is in a recursive map HiVE. */
	mc = hive_get_map_cache(sector, hc, &index);
	mc->flags |= flags;

	return mc->map[index];
}

static inline u64 hive_map_sector(u64 sector, struct hive_config *hc)
{
	return __hive_map_sector(sector, 0, hc);
}

/* Same as above -- cache hints are disabled at the moment. */
static inline u64 hive_map_sector_w_cache_hint(u64 sector, struct hive_config *hc)
{
	return __hive_map_sector(sector, HIVE_CACHE_HINT, hc);
}

/* Updates the forward map for the given sector. */
static void __hive_update_map(u64 sector, u64 new_sector, u32 flags, struct hive_config *hc)
{
	struct hive_map_cache *mc;
	u64 index;

	/* If map is in memory, update it. */
	if (hc->map) {
		hc->map[sector] = new_sector;
		return;
	}

	/* Otherwise, read from map HiVE and cache. */
	mc = hive_get_map_cache(sector, hc, &index);
	mc->map[index] = new_sector;
	mc->flags |= flags;
}

static inline void hive_update_map(u64 sector, u64 new_sector, struct hive_config *hc)
{
	__hive_update_map(sector, new_sector, HIVE_DIRTY, hc);
}

/* Same as above -- cache hints are disabled at the moment. */
static inline void hive_update_map_w_cache_hint(u64 sector, u64 new_sector, struct hive_config *hc)
{
	__hive_update_map(sector, new_sector, HIVE_DIRTY | HIVE_CACHE_HINT, hc);
}


/*
 * Meta block functions. Meta blocks contain reverse map entries and IVs for the
 * corresponding data sector.
 */

/* Returns the number of meta blocks accessed in a meta cache. */
static inline unsigned hive_decrypted_meta_count(struct hive_meta_cache *mc)
{
	return hive_bit_count(mc->is_decrypted, HIVE_META_PER_SECTOR);
}

/* Finds and returns the meta cache corresponding to the given sector. */
static struct hive_meta_cache *hive_find_meta_cache(u64 sector, struct hive_config *hc)
{
	u64 i;

	for (i = 0; i < hc->res->meta_cache_count; ++i) {
		if (hc->res->meta_cache[i]->sector == sector)
			return hc->res->meta_cache[i];
	}

	return NULL;
}

/* Caches the given meta block sector. */
static struct hive_meta_cache *hive_cache_meta(u64 sector, char *data, struct hive_config *hc)
{
	struct hive_meta_cache *mc;

	mc = hive_allocate_meta_cache(hc);
	mc->sector = sector;
	mc->data = data;
	hive_clear_bitset(mc->is_decrypted, HIVE_META_PER_SECTOR);
	hive_clear_bitset(mc->who_decrypted, HIVE_META_PER_SECTOR);

	hc->res->meta_cache[hc->res->meta_cache_count++] = mc;

	return mc;
}

/* Returns a previously cached meta entry. */
static struct hive_meta *__hive_get_meta(struct hive_meta_cache *mc, u64 index, struct hive_config *hc)
{
	struct hive_meta *m;
	u64 world;

	m = ((struct hive_meta *) mc->data) + index;

	/* If the meta block has not been decrypted yet, do it.*/
	if (!hive_test_bit(mc->is_decrypted, index)) {
		hive_do_crypto_from_buffer(((char *) m) + HIVE_IV_LEN, sizeof(m->r_map) + sizeof(m->world),
					m->iv, HIVE_DECRYPT, hc);
		hive_set_bit(mc->is_decrypted, index);
		hive_set_bit_to_value(mc->who_decrypted, index, (u8) hc->world);
	} else {
		/* Otherwise, if the meta block has been decrypted, see whodunit. */
		world = hive_test_bit(mc->who_decrypted, index) ? HIVE_WORLD_1 : HIVE_WORLD_0;

		/*
		 * If it was a different world, revert to the ciphertext, and
		 * then decrypt with this world's key.
		 */
		if (hc->world != world) {
			hive_do_crypto_from_buffer(((char *) m) + HIVE_IV_LEN, sizeof(m->r_map) + sizeof(m->world),
						m->iv, HIVE_ENCRYPT, hc->worlds[world]);

			hive_do_crypto_from_buffer(((char *) m) + HIVE_IV_LEN, sizeof(m->r_map) + sizeof(m->world),
						m->iv, HIVE_DECRYPT, hc);
			hive_set_bit_to_value(mc->who_decrypted, index, (u8) hc->world);
		}
	}

	return m;
}

/*
 * Returns a meta block. If the meta sector containing the block has not been
 * cached, first reads it from disk and caches it.
 */
static struct hive_meta *hive_get_meta(u64 sector, struct hive_config *hc)
{
	struct hive_meta_cache *mc;
	struct hive_meta *m;
	u64 meta_sector;
	char *data;
	u64 index;

	meta_sector = hc->res->meta_start + (sector / HIVE_META_PER_SECTOR);
	index = sector % HIVE_META_PER_SECTOR;
	mc = hive_find_meta_cache(meta_sector, hc);

	/* Read from disk if not already cached. */
	if (!mc) {
		data = hive_rw_sector(meta_sector, READ, NULL, hc);
		mc = hive_cache_meta(meta_sector, data, hc);
	}

	/* Decrypt and return the block. */
	m = __hive_get_meta(mc, index, hc);

	return m;
}

/* Encrypts and writes a meta cache back to disk. */
static void hive_write_meta(struct hive_meta_cache *mc, struct hive_config *hc)
{
	unsigned i;
	struct hive_meta *m;
	u64 who_encrypts;

	/* First encrypt all the decrypted meta blocks in the sector. */
	for (i = 0; i < HIVE_META_PER_SECTOR; ++i) {
		if (hive_test_bit(mc->is_decrypted, i)) {
			m = ((struct hive_meta *) mc->data) + i;

			/*
			 * If the meta block has an owner, i.e., we know which
			 * world uses the data sector, use that world's key for
			 * encryption.
			 */
			if (m->world < HIVE_MAX_WORLDS) {
				who_encrypts = m->world;
			} else { /* Otherwise, whoever decrypted the block encrypts it back. */
				who_encrypts = hive_test_bit(mc->who_decrypted, i) ? HIVE_WORLD_1 : HIVE_WORLD_0;
			}

			hive_do_crypto_from_buffer(((char *) m) + HIVE_IV_LEN, sizeof(m->r_map) + sizeof(m->world),
						m->iv, HIVE_ENCRYPT, hc->worlds[who_encrypts]);
		}
	}
	hive_rw_sector(mc->sector, WRITE, mc->data, hc);

	hive_free_sector(mc->data, hc);
	hive_free_meta_cache(mc, hc);
}

/* Evicts the meta block caches and writes them to disk. */
static void hive_evict_meta_cache(struct hive_config *hc)
{
	u64 i;

	for (i = 0; i < hc->res->meta_cache_count; ++i)
		hive_write_meta(hc->res->meta_cache[i], hc);

	hc->res->meta_cache_count = 0;
}

/* Returns the IV corresponding to a data sector. */
static void hive_get_iv(u64 sector, u8 *iv, struct hive_config *hc)
{
	struct hive_meta *m;

	m = hive_get_meta(sector, hc);
	memcpy(iv, m->iv, HIVE_IV_LEN);
}

/* Sets the IV corresponding to a data sector. */
static void hive_set_iv(u64 sector, u8 *iv, struct hive_config *hc)
{
	struct hive_meta *m;

	m = hive_get_meta(sector, hc);
	memcpy(m->iv, iv, HIVE_IV_LEN);
}

/* Returns the reverse map entry corresponding to a data sector. */
static u64 hive_reverse_map_sector(u64 sector, struct hive_config *hc, u64 *world)
{
	u64 r_sector;
	struct hive_meta *m;

	m = hive_get_meta(sector, hc);
	r_sector = m->r_map;
	if (world)
		*world = m->world;

	return r_sector;
}

/* Updates the reverse map entry corresponding to a data sector. */
static void hive_update_reverse_map(u64 sector, u64 new_sector, struct hive_config *hc)
{
	struct hive_meta *m;

	m = hive_get_meta(sector, hc);

	m->r_map = new_sector;
	m->world = hc->world;
}

/*
 * Called after overwriting a meta block with random data to indicate that we do
 * not need to encrypt it before writing.
 */
static void hive_set_meta_randomized(u64 sector, struct hive_config *hc)
{
	struct hive_meta_cache *mc;
	u64 meta_sector;
	u64 index;

	meta_sector = hc->res->meta_start + (sector / HIVE_META_PER_SECTOR);
	index = sector % HIVE_META_PER_SECTOR;
	mc = hive_find_meta_cache(meta_sector, hc);

	if (!mc) {
		DMCRIT("Meta not cached, cannot set randomized!");
		return;
	}

	hive_clear_bit(mc->is_decrypted, index);
}

/* Overwrites a meta block with random data. */
static void hive_randomize_meta(u64 sector, struct hive_config *hc)
{
	struct hive_meta *m;

	m = hive_get_meta(sector, hc);
	hive_fill_random_meta(m, hc);
	hive_set_meta_randomized(sector, hc);
}


/*
 * Stash functions.
 */

/* Checks whether a given sector is in stash, and returns it if found. */
static struct hive_stash *hive_is_in_stash(u64 sector, struct hive_config *hc)
{
	unsigned i;
	struct hive_stash *hs[HIVE_STASH_LEN];

	kfifo_out_peek(&hc->stash, hs, kfifo_len(&hc->stash));

	for (i = 0; i < kfifo_len(&hc->stash); ++i) {
		if (hs[i]->sector == sector)
			return hs[i];
	}

	return NULL;
}

/* Copies data from stash to page, or page to stash, according to @dir. */
static void __hive_copy_stash(struct page *p, u64 offset, struct hive_stash *hs, int dir)
{
	char *data;

	data = ((char *) kmap(p)) + offset;

	if (dir == HIVE_STASH_TO_PAGE)
		memcpy(data, hs->data, HIVE_SECTOR);
	else if (dir == HIVE_PAGE_TO_STASH)
		memcpy(hs->data, data, HIVE_SECTOR);

	kunmap(p);
}

/* Copies a page into an existing stash entry. */
static inline void hive_copy_page_to_stash(struct page *p, u64 offset, struct hive_stash *hs)
{
	__hive_copy_stash(p, offset, hs, HIVE_PAGE_TO_STASH);
}

/* Copies stashed data into a memory page. */
static inline void hive_copy_stash_to_page(struct page *p, u64 offset, struct hive_stash *hs)
{
	__hive_copy_stash(p, offset, hs, HIVE_STASH_TO_PAGE);
}

/* Enqueues the given data buffer in the corresponding world's stash. */
static void __hive_enqueue_stash(char *data, u64 hive_sector, struct hive_config *hc)
{
	struct hive_stash *hs;

	/* This should not happen. */
	if (unlikely(!kfifo_avail(&hc->stash)))
		DMCRIT("Out of stash in world %llu !", hc->world);

	hs = hive_allocate_stash(hc);
	memcpy(hs->data, data, HIVE_SECTOR);
	hs->sector = hive_sector;

	kfifo_in(&hc->stash, &hs, 1);
}

/* Stashes the contents of the given page. */
static void hive_enqueue_stash_from_page(struct page *p, u64 offset, u64 hive_sector, struct hive_config *hc)
{
	char *data;

	data = ((char *) kmap(p)) + offset;
	__hive_enqueue_stash(data, hive_sector, hc);

	/*
	 * Invalidate the map of the hive sector to indicate that the
	 * corresponding physical sector is now free.
	 */
	hive_update_map(hive_sector, HIVE_INVALID_SECTOR, hc);

	kunmap(p);
}

/* Reads and stashes the contents of a given disk sector. */
static void hive_enqueue_stash_from_disk(u64 disk_sector, u64 hive_sector, struct hive_config *hc)
{
	char *data;
	u8 iv[HIVE_IV_LEN];

	data = hive_rw_sector(hc->res->data_start + disk_sector, READ, NULL, hc);

	/* Must first decrypt the data. */
	hive_get_iv(disk_sector, iv, hc);
	hive_decrypt_sector_buffer(data, iv, hc);

	__hive_enqueue_stash(data, hive_sector, hc);

	/*
	 * Invalidate the map of the hive sector to indicate that the
	 * corresponding physical sector is now free.
	 */
	hive_update_map(hive_sector, HIVE_INVALID_SECTOR, hc);

	hive_free_sector(data, hc);
}

/* Writes the next stashed data in the given world to disk. */
static void __hive_write_from_stash(u64 to, struct hive_io_work *w, struct hive_config *hc)
{
	struct bio *bio;
	struct hive_stash *hs;
	struct page *p;
	u64 sector;

	/* This should not happen. */
	if (unlikely(kfifo_out(&hc->stash, &hs, 1) != 1))
		DMCRIT("Cannot dequeue in world %llu !", hc->world);

	/*
	 * Encrypt the data and convert the stash back to a page struct (Stashes
	 * are actually memory pages, see the trick at stash creation.)
	 */
	sector = hs->sector;
	hive_encrypt_sector_random_iv(to, hs->data, hc);
	p = virt_to_page(hs->data);
	hive_free_stash_keep_page(hs, hc);

	/* We will write the stash asynchronously. */
	bio = hive_allocate_bio(hc);
	bio->bi_sector = (hc->res->data_start + to) * HIVE_SECTOR_SCALE;
	bio->bi_private = w;
	bio->bi_end_io = hive_write_endio;
	bio->bi_bdev = hc->worlds[0]->real_dev->bdev;
	bio->bi_rw |= REQ_WRITE;
	bio_add_page(bio, p, HIVE_SECTOR, 0);

	/* Remap sector. */
	hive_update_reverse_map(to, sector, hc);
	/* hive_update_map_w_cache_hint(sector, to, hc); */ /* Disabled. */
	hive_update_map(sector, to, hc);

	/*
	 * Get an additional reference to the original bio before starting to
	 * write. We don't want to signal the original I/O for completion before
	 * the stash is written.
	 */
	bio_get(w->bio);
	submit_bio(HIVE_BIO_FLAGS, bio);
}

/* Writes the next stashed data to disk. World 0's stash has higher priority. */
static int hive_write_from_stash(u64 to, struct hive_io_work *w, struct hive_config *hc)
{
	if (kfifo_len(&hc->worlds[HIVE_WORLD_0]->stash)) {
		__hive_write_from_stash(to, w, hc->worlds[HIVE_WORLD_0]);
		return HIVE_WORLD_0;
	} else if (!hc->res->is_single_world && kfifo_len(&hc->worlds[HIVE_WORLD_1]->stash)) {
		__hive_write_from_stash(to, w, hc->worlds[HIVE_WORLD_1]);
		return HIVE_WORLD_1;
	}

	/* Stashes are empty, nothing written. */
	return HIVE_INVALID_WORLD;
}


/*
 * I/O functions.
 */

/*
 * Called on original non-split bios whenever a HiVE I/O operation is
 * finished. Ends the original bio if everything else is finished.
 */
static void hive_bio_put(struct bio *bio, struct hive_io_work *w, int free_work)
{
	if (atomic_dec_return(&bio->bi_cnt) == 1) {

		/* We can now dispatch the next I/O operation. */
		up(&w->hc->res->io_lock);

		if (free_work)
			hive_free_io_work(w, w->hc);

		bio_endio(bio, 0);
	}
}

/* End callbacks for HiVE-generated bios. */
static void __hive_write_endio(struct bio *bio, struct bio *bio_org, int error, struct hive_io_work *w, int free_work)
{
	if (unlikely(!bio_flagged(bio, BIO_UPTODATE) && !error))
		DMCRIT("WRITE ERROR, sector: %u", (unsigned) bio->bi_sector);

	hive_free_page(bio->bi_io_vec[0].bv_page, w->hc);
	bio->bi_io_vec[0].bv_page = NULL;

	hive_bio_put(bio_org, w, free_work);
	bio_put(bio);
}

static void hive_write_endio_no_free(struct bio *bio, int error)
{
	struct hive_io_work *w = (struct hive_io_work *) bio->bi_private;
	struct bio *bio_org = w->bio;

	__hive_write_endio(bio, bio_org, error, w, 0);
}

static void hive_write_endio(struct bio *bio, int error)
{
	struct hive_io_work *w = (struct hive_io_work *) bio->bi_private;
	struct bio *bio_org = w->bio;

	__hive_write_endio(bio, bio_org, error, w, 1);
}

static void hive_read_endio(struct bio *bio, int error)
{
	struct hive_crypto_work *cw = (struct hive_crypto_work *) bio->bi_private;

	if (unlikely(!bio_flagged(bio, BIO_UPTODATE) && !error))
		DMCRIT("READ ERROR, sector: %u",  (unsigned) bio->bi_sector);

	hive_queue_crypto(cw);
}

/* Returns whether the given data sector is free in the given world. */
static inline int hive_is_sector_free(u64 sector, struct hive_config *hc)
{
	u64 r_world;
	u64 r_sector;

	r_sector = hive_reverse_map_sector(sector, hc, &r_world);

	if (r_sector >= hc->world_len || r_world >= HIVE_MAX_WORLDS) {
		/*
		 * Meta block decryption was not successful. Sector must be free
		 * in this world.
		 */
		return 1;
	}

	r_sector = hive_map_sector(r_sector, hc);

	return !(r_world == hc->world && sector == r_sector);
}

/* Fills the given sector with random data. */
static void hive_randomize_sector(u64 sector, struct hive_io_work *w)
{
	struct bio *bio;
	struct page *p;
	char *data;

	bio = hive_allocate_bio(w->hc);
	p = hive_allocate_page(w->hc);

	data = (char *) kmap(p);
	hive_fill_random_sector(data, w->hc);
	kunmap(p);

	/* Asynchronous write. */
	bio->bi_private = w;
	bio->bi_end_io = hive_write_endio_no_free;
	bio->bi_bdev = w->hc->real_dev->bdev;
	bio->bi_rw |= REQ_WRITE;
	bio->bi_sector = (w->hc->res->data_start + sector) * HIVE_SECTOR_SCALE;
	bio_add_page(bio, p, HIVE_SECTOR, 0);

	/* Also randomize the corresponding meta block. */
	hive_randomize_meta(sector, w->hc);

	/*
	 * Get a reference to the original bio so that we don't signal its
	 * completion until randomization is complete.
	 */
	bio_get(w->bio);
	submit_bio(HIVE_BIO_FLAGS, bio);
}

/* No longer necessary.*/
/*
static void hive_submit_flush(struct hive_config *hc)
{
	struct bio *b;

	b = hive_allocate_bio(hc);
	b->bi_bdev = hc->real_dev->bdev;
	submit_bio_wait(WRITE | REQ_FLUSH, b);
	bio_put(b);
}
*/

/* Writes to K randomly picked sectors. */
static void __hive_do_write(struct hive_io_work *w)
{
	u64 write_sectors[HIVE_K];
	struct hive_config *hc;
	u64 hive_sector;
	u64 world;
	unsigned k;

	hc = w->hc;

	/* Get the sectors. */
	hive_get_random_sectors(HIVE_K, write_sectors, hc);

	for (k = 0; k < HIVE_K; ++k) {

		/* First check if any world owns the sector. */
		world = HIVE_INVALID_WORLD;
		if (!hive_is_sector_free(write_sectors[k], hc->worlds[HIVE_WORLD_0])) {
			world = HIVE_WORLD_0;
		} else if (!hc->res->is_single_world &&
			!hive_is_sector_free(write_sectors[k], hc->worlds[HIVE_WORLD_1])) {
			world = HIVE_WORLD_1;
		}

		/* If so, read from disk and stash it. */
		if (world != HIVE_INVALID_WORLD) {
			hive_sector = hive_reverse_map_sector(write_sectors[k], hc->worlds[world], &world);
			hive_enqueue_stash_from_disk(write_sectors[k], hive_sector, hc->worlds[world]);
		}
	}

	/*
	 * Write to the selected K sectors. If stash is empty, write random
	 * data.
	 */
	for (k = 0; k < HIVE_K; ++k) {
		if (hive_write_from_stash(write_sectors[k], w, hc) == HIVE_INVALID_WORLD)
			hive_randomize_sector(write_sectors[k], w);
	}
}

/* Writes operation. */
static void hive_do_write(struct hive_io_work *w)
{
	unsigned i;
	unsigned j;
	struct bio *bio;
	struct hive_config *hc;
	struct hive_stash *hs;
	u64 cur_sector;

	bio = w->bio;
	hc = w->hc;
	cur_sector = bio->bi_sector / HIVE_SECTOR_SCALE;

	/*
	 * We cannot dispatch multiple I/O works at the same time since they may
	 * access the same "random" blocks. Due to our blocking on heavy I/O, no
	 * actual performance is lost with this lock.
	 *
	 * BUT, this WILL slow down simultaneous World 0 & World 1 I/O.
	 */
	down(&hc->res->io_lock);

	/* Hold a reference to the original bio first. */
	bio_get(bio);

	/* Split the original bio into single-sector chunks and process them. */
	for (i = 0; i < bio->bi_vcnt; ++i) {
		for (j = 0; j < bio->bi_io_vec[i].bv_len; j += HIVE_SECTOR) {

			/*
			 * If this sector is already in stash, simply replace
			 * the stashed data with the new data.
			 */
			hs = hive_is_in_stash(cur_sector, hc);
			if (hs) {
				hive_copy_page_to_stash(bio->bi_io_vec[i].bv_page,
							bio->bi_io_vec[i].bv_offset + j, hs);
			} else {
				/* Otherwise, stash the new data. */
				hive_enqueue_stash_from_page(bio->bi_io_vec[i].bv_page,
							bio->bi_io_vec[i].bv_offset + j, cur_sector, hc);
			}

			/* Do K writes. */
			__hive_do_write(w);

			++cur_sector;

			hive_evict_meta_cache(hc);
			/* hive_evict_map_cache(hc); */ /* Disabled. */
			hive_evict_map_cache_all(hc);

			/* No longer necessary. */
			/* hive_submit_flush(hc); */
		}
	}
	/* hive_evict_map_cache_all(hc); */ /*
					      * Disabled. Not necessary without
					      * the cache hints.
					      */

	hive_bio_put(bio, w, 1); /* Release the extra reference. */
}

/* Read operation. */
static void hive_do_read(struct hive_io_work *w)
{
	unsigned i;
	unsigned j;
	struct bio *bio_new;
	struct bio *bio_org;
	struct hive_config *hc;
	struct hive_stash *hs;
	u64 cur_sector;
	u64 mapped_sector;
	struct hive_crypto_work *cw;

	bio_org = w->bio;
	hc = w->hc;
	cur_sector = bio_org->bi_sector / HIVE_SECTOR_SCALE;


	/*
	 * We cannot dispatch multiple I/O works at the same time since they may
	 * access the same "random" blocks. Due to our blocking on heavy I/O, no
	 * actual performance is lost with this lock.
	 *
	 * BUT, this WILL slow down simultaneous World 0 & World 1 I/O.
	 */
	down(&hc->res->io_lock);

	/* Hold a reference to the original bio first. */
	bio_get(bio_org);

	/* Split the original bio into single-sector chunks and process them. */
	for (i = 0; i < bio_org->bi_vcnt; ++i) {
		for (j = 0; j < bio_org->bi_io_vec[i].bv_len; j += HIVE_SECTOR) {

			/*
			 * Perform the K writes only if operating in World 0. Do
			 * this before doing the actual read operation -- the K
			 * writes can relocate or stash the sector to be read.
			 */
			if (hc->world == HIVE_WORLD_0)
				__hive_do_write(w);

			/*
			 * If this sector is already in stash, simply copy to
			 * the user the stash contents.
			 */
			hs = hive_is_in_stash(cur_sector, hc);
			if (hs) {
				hive_copy_stash_to_page(bio_org->bi_io_vec[i].bv_page,
							bio_org->bi_io_vec[i].bv_offset + j, hs);
			} else {
				/* mapped_sector = hive_map_sector_w_cache_hint(cur_sector, hc); */ /* Disabled. */
				mapped_sector = hive_map_sector(cur_sector, hc);

				/* Otherwise, do an actual asynchronous read. */
				bio_new = hive_allocate_bio(hc);
				cw = hive_allocate_crypto_work(hc);

				bio_new->bi_private = cw;
				bio_new->bi_end_io = hive_read_endio;
				bio_new->bi_bdev = hc->real_dev->bdev;
				bio_new->bi_rw = bio_org->bi_rw;
				bio_new->bi_sector = (hc->res->data_start + mapped_sector) * HIVE_SECTOR_SCALE;

				cw->io = w;
				cw->bio = bio_new;
				hive_get_iv(mapped_sector, cw->iv, hc);

				bio_add_page(bio_new, bio_org->bi_io_vec[i].bv_page, HIVE_SECTOR,
					bio_org->bi_io_vec[i].bv_offset + j);

				/*
				 * Important - Get an additional reference to
				 * the original bio before starting to write.
				 */
				bio_get(bio_org);
				submit_bio(HIVE_BIO_FLAGS, bio_new);
			}

			++cur_sector;

			hive_evict_meta_cache(hc);
			/* hive_evict_map_cache(hc); */ /* Disabled. */
			hive_evict_map_cache_all(hc);

			/* No longer necessary. */
			/* hive_submit_flush(hc); */
		}
	}
	/* hive_evict_map_cache_all(hc); */ /*
					      * Disabled. Not necessary without
					      * the cache hints.
					      */

	hive_bio_put(bio_org, w, 1); /* Release the extra reference. */
}


/* Bottom-half entry, redirects the original bio to the correct handler. */
static void hive_do_io(struct work_struct *work)
{
	struct hive_io_work *w = container_of(work, struct hive_io_work, work);

	if (bio_data_dir(w->bio) == WRITE)
		hive_do_write(w);
	else
		hive_do_read(w);
}

/* Bottom-half entry, decrypts a bio after a read operation. */
static void hive_do_crypto(struct work_struct *work)
{
	struct hive_crypto_work *cw = container_of(work, struct hive_crypto_work, work);
	struct hive_io_work *io_w = cw->io;
	struct bio *bio = cw->bio;
	struct bio *bio_org = io_w->bio;

	hive_decrypt_sector_page(bio->bi_io_vec[0].bv_page, bio->bi_io_vec[0].bv_offset, cw->iv, cw->hc);

	hive_free_crypto_work(cw, io_w->hc);
	hive_bio_put(bio_org, io_w, 1);
	bio_put(bio);
}

/* Adds a bio to the IO queue to be taken care of later. */
static void hive_queue_io(struct hive_io_work *w)
{
	INIT_WORK(&w->work, hive_do_io);
	queue_work(w->hc->res->io_queue, &w->work);
}

/* Adds a read bio to the crypto queue to be decrypted later. */
static void hive_queue_crypto(struct hive_crypto_work *w)
{
	INIT_WORK(&w->work, hive_do_crypto);
	queue_work(w->hc->res->crypto_queue, &w->work);
}


/*
 * Initializers, constructors & destructors.
 */

/* Reads and returns the HiVE headers from disk. */
static inline struct hive_header *hive_read_headers(u64 start, struct hive_config *hc)
{
	return (struct hive_header *) hive_rw_sector(start, READ, NULL, hc);
}

/* Writes the HiVE headers back to disk. */
static inline void hive_write_headers(struct hive_header *h, u64 start, struct hive_config *hc)
{
	hive_rw_sector(start, WRITE, (char *) h, hc);
}

/* If using an in-memory forward map, reads and returns the map from disk. */
static u64 *hive_read_map(u64 start, u64 len, u8 *iv, struct hive_config *hc)
{
	u64 i;
	char *data;
	char *map;

	/* Could be big. */
	map = vmalloc(len * HIVE_SECTOR);

	data = hive_rw_sector(start, READ, NULL, hc);
	hive_do_crypto_from_buffer(data, HIVE_SECTOR, iv, HIVE_DECRYPT, hc);
	memcpy(map, data, HIVE_SECTOR);
	hive_free_sector(data, hc);

	for (i = 1; i < len; ++i) {
		data = hive_rw_sector(start + i, READ, NULL, hc);
		hive_do_crypto_from_buffer(data, HIVE_SECTOR, NULL, HIVE_DECRYPT, hc);
		memcpy(map + (i * HIVE_SECTOR), data, HIVE_SECTOR);
		hive_free_sector(data, hc);
	}

	return (u64 *) map;
}

/* Writes the in-memory forward map back to disk. */
static void hive_write_map(char *map, u64 start, u64 len, u8 *iv, struct hive_config *hc)
{
	struct page *p;
	char *data;
	u64 i;

	p = hive_allocate_page(hc);
	data = kmap(p);

	memcpy(data, map, HIVE_SECTOR);
	hive_do_crypto_from_buffer(data, HIVE_SECTOR, iv, HIVE_ENCRYPT, hc);
	hive_rw_sector(start, WRITE, data, hc);

	for (i = 1; i < len; ++i) {
		memcpy(data, map + (i * HIVE_SECTOR), HIVE_SECTOR);
		hive_do_crypto_from_buffer(data, HIVE_SECTOR, NULL, HIVE_ENCRYPT, hc);
		hive_rw_sector(start + i, WRITE, data, hc);
	}
	kunmap(p);
	hive_free_page(p, hc);
}

/* Reads the saved stash entries from disk. */
static void hive_read_stash(u64 start, u64 map_start, u8 *iv, struct hive_config *hc)
{
	u64 *map;
	u8 *stash;
	unsigned i;

	/*
	 * We have a simple stash map that indicates how many stash entries are
	 * stored on disk. This must fit in a single block.
	 */
	map = (u64 *) hive_rw_sector(map_start, READ, NULL, hc);
	hive_do_crypto_from_buffer((char *) map, HIVE_SECTOR, iv, HIVE_DECRYPT, hc);

	for (i = 0; i < HIVE_STASH_LEN; ++i) {
		if (map[i] >= hc->world_len) {
			DMCRIT("Read %u stash entries.", i);
			break;
		}

		stash = hive_rw_sector(start + i, READ, NULL, hc);
		hive_do_crypto_from_buffer(stash, HIVE_SECTOR, NULL, HIVE_DECRYPT, hc);
		__hive_enqueue_stash(stash, map[i], hc);
		hive_free_sector(stash, hc);
	}

	hive_free_sector((char *) map, hc);
}

/* Writes the contents of the stashes to disk. */
static void hive_write_stash(u64 start, u64 map_start, u8 *iv, struct hive_config *hc)
{
	struct hive_stash *hs;
	u8 *stash;
	u64 *map;
	unsigned i;
	unsigned count;
	u8 *buf;
	struct page *p;

	/*
	 * Again, We create a simple stash map that indicates how many stash
	 * entries are stored on disk, and write it too.
	 */
	map = kmalloc(HIVE_SECTOR, GFP_KERNEL);
	stash = vmalloc(HIVE_STASH_LEN * HIVE_SECTOR);

	for (i = 0; !kfifo_is_empty(&hc->stash); ++i) {
		kfifo_out(&hc->stash, &hs, 1);
		map[i] = hs->sector;
		memcpy(stash + (i * HIVE_SECTOR), hs->data, HIVE_SECTOR);
		hive_free_stash(hs, hc);
	}
	count = i;

	for (; i < HIVE_STASH_LEN; ++i) {
		map[i] = HIVE_INVALID_STASH;
		hive_fill_random_sector(stash + (i * HIVE_SECTOR), hc);
	}

	p = hive_allocate_page(hc);
	buf = kmap(p);

	memcpy(buf, map, HIVE_SECTOR);
	hive_do_crypto_from_buffer(buf, HIVE_SECTOR, iv, HIVE_ENCRYPT, hc);
	hive_rw_sector(map_start, WRITE, buf, hc);

	for (i = 0; i < count; ++i) {
		memcpy(buf, stash + (i * HIVE_SECTOR), HIVE_SECTOR);
		hive_do_crypto_from_buffer(buf, HIVE_SECTOR, NULL, HIVE_ENCRYPT, hc);
		hive_rw_sector(start + i, WRITE, buf, hc);
	}

	for (; i < HIVE_STASH_LEN; ++i) {
		memcpy(buf, stash + (i * HIVE_SECTOR), HIVE_SECTOR);
		hive_rw_sector(start + i, WRITE, buf, hc);
	}

	DMCRIT("Written %u stash entries.", count);

	kunmap(p);
	hive_free_page(p, hc);
	vfree(stash);
	kfree(map);
}

/* Destroys the stashes on error. */
static void hive_destroy_stash(u64 worlds, struct hive_config *hc)
{
	struct hive_stash *hs;
	unsigned i;
	unsigned w;

	for (w = 0; w < worlds; ++w) {
		for (i = 0; !kfifo_is_empty(&hc->worlds[w]->stash); ++i) {
			kfifo_out(&hc->worlds[w]->stash, &hs, 1);
			hive_free_stash(hs, hc->worlds[w]);
		}
	}
}

/* Initializes and returns a crypto transform with the given key. */
static struct crypto_blkcipher *hive_init_tfm(u8 *key)
{
	struct crypto_blkcipher *tfm;

	tfm = crypto_alloc_blkcipher("cbc(aes)", 0, 0);

	if (IS_ERR(tfm)) {
		DMCRIT("Failed to allocate crypto transform");
		return NULL;
	}

	/* We won't need the key after this step. */
	if (crypto_blkcipher_setkey(tfm, key, HIVE_KEY_LEN)) {
		DMCRIT("Failed to set key");
		return NULL;
	}

	return tfm;
}

/* Random context ctr and dtr. */
#if (HIVE_PRNG == HIVE_PRNG_AESCTR)
/* Initializes the AES-CTR context. */
static int hive_init_rand_context(struct hive_rand_context *rand)
{
	u8 ctr[16];

	rand->buf = NULL;

	rand->tfm = crypto_alloc_blkcipher("ctr(aes)", 0, 0);
	if (IS_ERR(rand->tfm))
		return HIVE_ERROR;
	/* Key will be set later when filling the buffer. */

	/* Set random counter. */
	hive_get_random_bytes_kernel(ctr, 16);
	crypto_blkcipher_set_iv(rand->tfm, ctr, 16);

	rand->max_byte = HIVE_PRNG_AESCTR_REFRESH_LEN / HIVE_PRNG_AESCTR_CHUNK_LEN;
	rand->max_chunk = HIVE_PRNG_AESCTR_CHUNK_LEN;

	rand->buf = kmalloc(rand->max_byte, GFP_KERNEL);
	while (!rand->buf) {
		DMCRIT("Cannot allocate memory for AES-CTR, %llu bytes", rand->max_byte);

		if (rand->max_byte <= PAGE_SIZE) {
			DMCRIT("Bailing out");
			crypto_free_blkcipher(rand->tfm);
			rand->tfm = NULL;
			return HIVE_ERROR;
		}

		rand->max_byte = rand->max_byte << 2;
		rand->max_chunk = rand->max_chunk >> 2;

		DMCRIT("Increasing chunks to: %llu", rand->max_chunk);

		rand->buf = kmalloc(rand->max_byte, GFP_KERNEL);
	}

	/* Initialize to invalid values. */
	rand->cur_byte = rand->max_byte;
	rand->cur_chunk = rand->max_chunk;

	return HIVE_SUCCESS;
}

static void hive_destroy_rand_context(struct hive_rand_context *rand)
{
	if (rand->tfm)
		crypto_free_blkcipher(rand->tfm);

	kfree(rand->buf);
}

#elif (HIVE_PRNG == HIVE_PRNG_ARC4)
/* Initializes the ARC4 context. */
static int hive_init_rand_context(struct hive_rand_context *rand)
{
	u8 *S;
	unsigned i;
	unsigned j;
	unsigned k;

	u8 key[256];

	hive_get_random_bytes_kernel(key, 256);

	S = rand->S;
	for (i = 0; i < 256; ++i)
		S[i] = i;

	j = 0;
	for (i = 0; i < 256; ++i) {
		j = (j + S[i] + key[i]) & 0xff;
		swap(S[i], S[j]);
	}

	/* Discard the first 256 bytes for good measure. */
	i = 0;
	j = 0;
	for (k = 0; k < 256; ++k) {
		i = (i + 1) & 0xff;
		j = (j + S[i]) & 0xff;
		swap(S[i], S[j]);
	}

	rand->i = i;
	rand->j = j;

	memset(key, 0, 256);

	return HIVE_SUCCESS;
}

/* Nothing to destroy for ARC4. */
static void hive_destroy_rand_context(struct hive_rand_context *rand)
{
}
#endif

#ifndef HIVE_DUMB_RANDOM_SECTORS
/*
 * Initializes the random sector permutation. Only used if using the partial
 * Knuth-shuffle to get the random sectors. Can be ridiculously large, hence we
 * don't normally use it.
 */
static u64 *hive_init_random_list(u64 len)
{
	u64 i;

	u64 *list = kmalloc(len * sizeof(*list), GFP_KERNEL);
	if (!list) {
		DMCRIT("Cannot allocate memory! Now what?");
		return NULL;
	}

	for (i = 0; i < len; ++i)
		list[i] = i;

	return list;
}
#endif

/*
 * DM mapping function. This executes under make_generic_request(), and under
 * that functions submit_bio() does not actually submit anything until we
 * return. This makes synchronous I/O impossible in this function, and we need
 * to do it for meta block and sector mapping operations. Therefore here we only
 * set up a bottom-half to do the actual job later.
 */
static int hive_map_bio(struct dm_target *ti, struct bio *bio)
{
	struct hive_io_work *w;
	struct hive_config *hc = (struct hive_config *) ti->private;

	bio->bi_bdev = hc->real_dev->bdev;

	if (unlikely(bio->bi_rw & (REQ_FLUSH | REQ_DISCARD))) {
		bio_endio(bio, 0);
		return DM_MAPIO_SUBMITTED;
	}

	if (bio_sectors(bio)) {
		w = hive_allocate_io_work(bio, hc);
		hive_queue_io(w);
		return DM_MAPIO_SUBMITTED;
	}

	return DM_MAPIO_REMAPPED;
}

/*
 * Constructor. Called on world creation.
 */
static int hive_ctr(struct dm_target *ti, unsigned int argc, char **argv)
{
	struct hive_dev *hd;
	struct hive_dev *hd_map;
	struct hive_config *hc;
	struct hive_resource *res;
	u32 world;

	u64 start;
	u32 is_single_world;
	char dummy;
	u8 *hive_keys[HIVE_MAX_WORLDS];

	unsigned i;
	unsigned err_i;
	unsigned j;
	unsigned max_worlds;
	struct hive_header *headers;

	if (argc < 7) {
		ti->error = "Invalid argument count";
		return -EINVAL;
	}

	DMCRIT("Creating HiVE on %s", argv[0]);

	if (argv[4][0] == '1') {
		is_single_world = 1;
		max_worlds = 1;
	} else {
		is_single_world = 0;
		max_worlds = HIVE_MAX_WORLDS;
	}

	/* Lock everything until things are stable. */
	down(&hive_dev_lock);

	hd = hive_lookup_dev(argv[0], argv[1]);
	if (!hd) {
		hd = hive_create_dev(argv[0], argv[1], is_single_world);
		if (!hd)
			return -EINVAL;
	}

	if (hive_get_dev(hd, &world, ti) == HIVE_ERROR)
		return -EINVAL;

	hive_set_world_path(argv[2], hd);

	if (sscanf(argv[3], "%llu%c", &start, &dummy) != 1) {
		ti->error = "Invalid arguments";
		return -EINVAL;
	}

	hc = hd->hc[world];
	if (dm_get_device(ti, argv[0], dm_table_get_mode(ti->table), &hc->real_dev)) {
		ti->error = "Real device lookup failed";
		goto bad_put_hive_dev;
	}


	/*
	 * We do bulk of the work when creating the first world in a HiVE
	 * instance. This is necessary since World 0 also needs access to World 1's maps
	 * and metadata.
	 */
	if (world == 0) {
		/* Some initial setup. */
		for (i = 0; i < max_worlds; ++i) {
			for (j = 0; j < max_worlds; ++j)
				hd->hc[i]->worlds[j] = hd->hc[j];

			hd->hc[i]->world = i;

			hive_keys[i] = hive_hex_decode(argv[5 + i]);
		}

		res = kmalloc(sizeof(*res), GFP_KERNEL);
		memset(res, 0, sizeof(*res));

		hc->res = res;
		res->is_single_world = is_single_world;
		res->dev = hd;

		/* Create memory pools, work queues, locks... */
		res->bioset = bioset_create(HIVE_BIOSET_SIZE, 0);
		if (!res->bioset) {
			ti->error = "Could not create bioset.";
			goto bad_free_res_put_dev;
		}

		res->page_pool = mempool_create_page_pool(HIVE_PAGEPOOL_SIZE, 0);
		if (!res->page_pool) {
			ti->error = "Could not create page pool.";
			goto bad_free_bioset;
		}

		res->_io_work_pool = KMEM_CACHE(hive_io_work, 0);
		if (!res->_io_work_pool) {
			ti->error = "Could not create io cache.";
			goto bad_free_page_pool;
		}

		res->io_work_pool = mempool_create_slab_pool(HIVE_IO_WORKPOOL_SIZE, res->_io_work_pool);
		if (!res->io_work_pool) {
			ti->error = "Could not create io pool.";
			goto bad_free_io_work_cache;
		}

		res->_crypto_work_pool = KMEM_CACHE(hive_crypto_work, 0);
		if (!res->_crypto_work_pool) {
			ti->error = "Could not create crypto cache.";
			goto bad_free_io_work_pool;
		}

		res->crypto_work_pool = mempool_create_slab_pool(HIVE_CRYPTO_WORKPOOL_SIZE, res->_crypto_work_pool);
		if (!res->crypto_work_pool) {
			ti->error = "Could not create crypto pool.";
			goto bad_free_crypto_work_cache;
		}

		res->_meta_cache_pool = KMEM_CACHE(hive_meta_cache, 0);
		if (!res->_meta_cache_pool) {
			ti->error = "Could not create meta cache.";
			goto bad_free_crypto_work_pool;
		}

		res->meta_cache_pool = mempool_create_slab_pool(HIVE_METACACHEPOOL_SIZE, res->_meta_cache_pool);
		if (!res->meta_cache_pool) {
			ti->error = "Could not create meta pool.";
			goto bad_free_meta_cache_cache;
		}

		res->_map_cache_pool = KMEM_CACHE(hive_map_cache, 0);
		if (!res->_map_cache_pool) {
			ti->error = "Could not create map cache.";
			goto bad_free_meta_cache_pool;
		}

		res->map_cache_pool = mempool_create_slab_pool(HIVE_MAPCACHEPOOL_SIZE, res->_map_cache_pool);
		if (!res->map_cache_pool) {
			ti->error = "Could not create meta pool.";
			goto bad_free_map_cache_cache;
		}

		res->_stash_pool = KMEM_CACHE(hive_stash, 0);
		if (!res->_stash_pool) {
			ti->error = "Could not create stash cache.";
			goto bad_free_map_cache_pool;
		}

		res->stash_pool = mempool_create_slab_pool(HIVE_STASHPOOL_SIZE, res->_stash_pool);
		if (!res->stash_pool) {
			ti->error = "Could not create stash pool.";
			goto bad_free_stash_cache;
		}

		res->io_queue = create_singlethread_workqueue("hive_io");
		if (!res->io_queue) {
			ti->error = "Could not create io queue.";
			goto bad_free_stash_pool;
		}

		res->crypto_queue = create_singlethread_workqueue("hive_crypto");
		if (!res->crypto_queue) {
			ti->error = "Could not create crypto queue.";
			goto bad_free_io_queue;
		}

		sema_init(&res->io_lock, 1);
		sema_init(&res->crypto_lock, 1);

		if (hive_init_rand_context(&res->rand) == HIVE_ERROR) {
			ti->error = "Could not create random context.";
			goto bad_free_crypto_queue;
		}

		headers = hive_read_headers(start, hc);
		if (!headers) {
			ti->error = "Could not read headers.";
			goto bad_destroy_rand_context;
		}

#ifndef HIVE_DUMB_RANDOM_SECTORS
		res->random_list = hive_init_random_list(headers[0].data_len);
		if (!res->random_list) {
			ti->error = "Could not create random list.";
			goto bad_free_headers;
		}
#endif

		res->start = start;
		res->data_start = start + headers[0].data_start;
		res->data_len = headers[0].data_len;
		res->meta_start = start + headers[0].meta_start;

		for (i = 0; i < max_worlds; ++i) {
			hd->hc[i]->tfm = hive_init_tfm(hive_keys[i]);
			if (!hd->hc[i]->tfm) {
				ti->error = "Could not create crypto transform.";
				goto bad_free_random_list_and_tfm;
			}

			/* We don't need the keys anymore, wipe them clean. */
			memset(hive_keys[i], 0, HIVE_KEY_LEN);
			memset(argv[5 + i], 0, HIVE_KEY_LEN);
			kfree(hive_keys[i]);
		}

		for (i = 0; i < max_worlds; ++i) {
			hd->hc[i]->world_len = headers[i].world_len;
			hd->hc[i]->res = hd->hc[0]->res;
			INIT_KFIFO(hd->hc[i]->stash);

			if (headers[i].map_len) {
				hd->hc[i]->map = hive_read_map(start + headers[i].map_start,
								headers[i].map_len, headers[i].iv, hd->hc[i]);
			} else {
				hd->hc[i]->map = NULL;
			}
		}

		for (i = 0; i < max_worlds; ++i) {
			hive_read_stash(start + headers[i].stash_start,
					start + headers[i].stash_map_start, headers[i].iv, hd->hc[i]);
		}

		/*
		 * If this HiVE recursively uses another one for the forward
		 * map entries, link them together here.
		 */
		if (argc == 8) {
			DMCRIT("Map hive is %s", argv[7]);

			hd_map = hive_lookup_dev(argv[0], argv[7]);
			if (!hd_map) {
				DMCRIT("Cannot find map hive!");
				goto bad_destroy_stash;
			} else if (hd_map->worlds != max_worlds) {
				DMCRIT("Hive world number does not match map hive!");
				goto bad_destroy_stash;
			}

			hd->map_dev = hd_map;

			for (i = 0; i < max_worlds; ++i) {
				if (dm_get_device(ti, hd_map->world_paths[i],
							dm_table_get_mode(ti->table), &hd->hc[i]->map_dev)) {
					ti->error = "Map device lookup failed.";
					goto bad_put_map_dev;
				}
			}
		} else {
			hd->map_dev = NULL;
			for (i = 0; i < max_worlds; ++i)
				hd->hc[i]->map_dev = NULL;
		}

		hive_free_sector((char *) headers, hc);

		/* Change to noop scheduler. */
		elevator_change(hc->real_dev->bdev->bd_queue, "noop");
	}

	ti->num_discard_bios = 1;
	ti->private = hc;

	/* All set. Ready to let go of the lock. */
	up(&hive_dev_lock);

	DMCRIT("Success.");
	return 0;

/* Lots to clean up after an error. */
bad_put_map_dev:
	for (err_i = 0; err_i < i; ++err_i)
		dm_put_device(ti, hd->hc[err_i]->map_dev);
bad_destroy_stash:
	hive_destroy_stash(max_worlds, hc);
	i = max_worlds;
bad_free_random_list_and_tfm:
	for (err_i = 0; err_i < i; ++err_i)
		crypto_free_blkcipher(hd->hc[err_i]->tfm);
#ifndef HIVE_DUMB_RANDOM_SECTORS
	kfree(res->random_list);
bad_free_headers:
#endif
	hive_free_sector((char *) headers, hc);
bad_destroy_rand_context:
	hive_destroy_rand_context(&res->rand);
bad_free_crypto_queue:
	destroy_workqueue(res->crypto_queue);
bad_free_io_queue:
	destroy_workqueue(res->io_queue);
bad_free_stash_pool:
	mempool_destroy(res->stash_pool);
bad_free_stash_cache:
	kmem_cache_destroy(res->_stash_pool);
bad_free_map_cache_pool:
	mempool_destroy(res->map_cache_pool);
bad_free_map_cache_cache:
	kmem_cache_destroy(res->_map_cache_pool);
bad_free_meta_cache_pool:
	mempool_destroy(res->meta_cache_pool);
bad_free_meta_cache_cache:
	kmem_cache_destroy(res->_meta_cache_pool);
bad_free_crypto_work_pool:
	mempool_destroy(res->crypto_work_pool);
bad_free_crypto_work_cache:
	kmem_cache_destroy(res->_crypto_work_pool);
bad_free_io_work_pool:
	mempool_destroy(res->io_work_pool);
bad_free_io_work_cache:
	kmem_cache_destroy(res->_io_work_pool);
bad_free_page_pool:
	mempool_destroy(res->page_pool);
bad_free_bioset:
	bioset_free(res->bioset);
	for (i = 0; i < max_worlds; ++i)
		kfree(hive_keys[i]);
bad_free_res_put_dev:
	kfree(res);
	dm_put_device(ti, hc->real_dev);
bad_put_hive_dev:
	if (world == 0)
		hive_put_dev(hd);

	up(&hive_dev_lock);

	return -EINVAL;
}

/*
 * Destructor. Called on world destruction (This sounds so cool).
 */
static void hive_dtr(struct dm_target *ti)
{
	u8 *data;
	unsigned i;
	unsigned j;
	struct hive_header *headers;
	unsigned max_worlds = HIVE_MAX_WORLDS;
	struct hive_config *hc = (struct hive_config *) ti->private;
	struct hive_resource *res = hc->res;
	struct hive_dev *hd = res->dev;

	DMCRIT("Destroying.");

	/* Lock until everything is cleaned up. */
	down(&hive_dev_lock);

	/*
	 * The HiVE kernel module does not know/care about the
	 * creation/destruction order, which is tracked by the userspace
	 * tool. This should never happen unless the curious user plays around
	 * with the device-mapper ioctl interface directly.
	 */
	if (hd->worlds == 2 && hc->world == HIVE_WORLD_0)
		DMCRIT("Warning: Removing World 0 before World 1.");

	/*
	 * Similar to the constructor, we actually destroy things only after the
	 * last world has been removed.
	 */
	if (hd->worlds == 1) {

		if (res->is_single_world)
			max_worlds = 1;

		headers = hive_read_headers(res->start, hc);
		for (i = 0; i < HIVE_MAX_WORLDS; ++i) {
			hive_get_random_iv(headers[i].iv, hc);

			if (i < max_worlds) {
				if (hd->hc[i]->map) {
					hive_write_map((char *) hd->hc[i]->map,
							res->start + headers[i].map_start,
							headers[i].map_len, headers[i].iv, hd->hc[i]);
					vfree(hd->hc[i]->map);
				} else {
					dm_put_device(ti, hd->hc[i]->map_dev);
				}

				hive_write_stash(res->start + headers[i].stash_start,
						res->start + headers[i].stash_map_start,
						headers[i].iv, hd->hc[i]);
			} else {
				/*
				 * If operating in single-world mode, we
				 * overwrite the map and stash of the
				 * non-existent World 1 with random data.
				 */
				data = kmalloc(HIVE_SECTOR, GFP_KERNEL);

				if (hc->map) {
					for (j = 0; j < headers[i].map_len; ++j) {
						hive_fill_random_sector(data, hc);
						hive_rw_sector(res->start + headers[i].map_start + j, WRITE, data, hc);
					}
				}

				for (j = 0; j < HIVE_STASH_LEN; ++j) {
					hive_fill_random_sector(data, hc);
					hive_rw_sector(res->start + headers[i].stash_start + j, WRITE, data, hc);
				}

				hive_fill_random_sector(data, hc);
				hive_rw_sector(res->start + headers[i].stash_map_start, WRITE, data, hc);

				kfree(data);
			}
		}

		hive_write_headers(headers, res->start, hc);
		hive_free_sector((char *) headers, hc);

		for (i = 0; i < max_worlds; ++i)
			crypto_free_blkcipher(hd->hc[i]->tfm);

		hive_destroy_rand_context(&res->rand);

#ifndef HIVE_DUMB_RANDOM_SECTORS
		kfree(res->random_list);
#endif
		destroy_workqueue(res->crypto_queue);
		destroy_workqueue(res->io_queue);

		mempool_destroy(res->stash_pool);
		kmem_cache_destroy(res->_stash_pool);

		mempool_destroy(res->map_cache_pool);
		kmem_cache_destroy(res->_map_cache_pool);

		mempool_destroy(res->meta_cache_pool);
		kmem_cache_destroy(res->_meta_cache_pool);

		kmem_cache_destroy(res->_crypto_work_pool);
		mempool_destroy(res->crypto_work_pool);

		kmem_cache_destroy(res->_io_work_pool);
		mempool_destroy(res->io_work_pool);

		mempool_destroy(res->page_pool);

		bioset_free(res->bioset);

		kfree(res);
	}

	dm_put_device(ti, hc->real_dev);
	hive_put_dev(hd);

	up(&hive_dev_lock);
	DMCRIT("Success.");
}

/*
 * Terrible hack incoming: As far as I understand device-mapper does not provide
 * us with an easy way to change device limitations but instead computes sane
 * values according to the underlying block devices. Well, we need 4KB logical
 * blocks, so we lie about the block device geometry.
 */
static int hive_iterate_devices(struct dm_target *ti, iterate_devices_callout_fn fn, void *data)
{
	struct hive_config *hc = ti->private;

	if (fn == dm_set_device_limits) {
		struct queue_limits *q = data;
		q->logical_block_size = HIVE_SECTOR;
		q->physical_block_size = HIVE_SECTOR;
		q->io_min = HIVE_SECTOR;
	}
	return fn(ti, hc->real_dev, 0, ti->len, data);
}

/*
 * TODO: The commented out stuff is not absolutely necessary. Actually I am not
 * sure what HiVE has to offer for them. I'll implement them when I find a good
 * use for them.
 */
static struct target_type hive_target = {
	.name   = "hive",
	.version = {1, 0, 0},
	.module = THIS_MODULE,
	.ctr    = hive_ctr,
	.dtr    = hive_dtr,
	.map    = hive_map_bio,
	/* .status = hive_status, */
	/* .ioctl  = hive_ioctl, */
	/* .merge  = hive_merge, */
	.iterate_devices = hive_iterate_devices,
};

/* Module entry. */
static int __init dm_hive_init(void)
{
	int r = dm_register_target(&hive_target);

	if (r < 0)
		DMERR("dm_register failed %d", r);

	if (!proc_create(HIVE_PROC_FILE, 0, NULL, &hive_fops)) {
		DMERR("Cannot create proc file.");
		return -ENOMEM;
	}

	DMCRIT("HiVE loaded.");
	return r;
}

/* Module exit. */
static void __exit dm_hive_exit(void)
{
	remove_proc_entry(HIVE_PROC_FILE, NULL);
	dm_unregister_target(&hive_target);
	DMCRIT("HiVE unloaded.");
}

module_init(dm_hive_init);
module_exit(dm_hive_exit);

MODULE_AUTHOR("Kaan Onarlioglu <onarliog@ccs.neu.edu>");
MODULE_DESCRIPTION(DM_NAME " target for HiVE.");
MODULE_LICENSE("GPL");
